from __future__ import annotations

import json
from typing import Optional, Callable, Union, List, Dict, Tuple, Any, Set

from spacy.matcher import Matcher
from spacy.tokens import Doc, Span

from ..common.base_rule import BaseRule


# import warnings

# warnings.simplefilter("always")


class ConTextRule(BaseRule):
    """
    A ConTextRule defines a ConText modifier. ConTextRules are rules which define which spans are extracted as modifiers
    and how they behave, such as the phrase to be matched, the category/semantic class, the direction of the modifier in
    the text, and what types of target spans can be modified.
    """

    _ALLOWED_DIRECTIONS = (
        "FORWARD",
        "BACKWARD",
        "BIDIRECTIONAL",
        "TERMINATE",
        "PSEUDO",
    )
    _ALLOWED_KEYS = {
        "literal",
        "direction",
        "pattern",
        "category",
        "metadata",
        "allowed_types",
        "excluded_types",
        "max_targets",
        "max_scope",
    }

    def __init__(
        self,
        literal: str,
        category: str,
        pattern: Optional[Union[str, List[Dict[str, str]]]] = None,
        direction: str = "BIDIRECTIONAL",
        on_match: Optional[
            Callable[[Matcher, Doc, int, List[Tuple[int, int, int]]], Any]
        ] = None,
        on_modifies: Optional[Callable[[Span, Span, Span], bool]] = None,
        allowed_types: Optional[Set[str]] = None,
        excluded_types: Optional[Set[str]] = None,
        max_scope: Optional[int] = None,
        max_targets: Optional[int] = None,
        terminated_by: Optional[Set[str]] = None,
        metadata: Optional[Dict[Any, Any]] = None,
    ):
        """
        Creates a ConTextRule object.

        The primary arguments of `literal` `category`, and `direction` define the span of text to be matched, the
        semantic category, and the direction within the sentence in which the modifier operates.
        Other arguments specify additional custom logic such as:
            - Additional control over what text can be matched as a modifier (pattern and on_match)
            - Which types of targets can be modified (allowed_types, excluded_types)
            - The scope size and number of targets that a modifier can modify (max_targets, max_scope)
            - Other logic for terminating a span or for allowing a modifier to modify a target (on_modifies,
            terminated_by)

        Args:
            literal: The string representation of a concept. If `pattern` is None, this string will be lower-cased and
                matched to the lower-case string. If `pattern` is not None, this argument will not be used for matching
                but can be used as a reference as the rule name.
            category: The semantic class of the matched span. This corresponds to the `label_` attribute of an entity.
            pattern: A list or string to use as a spaCy pattern rather than `literal`. If a list, will use spaCy
                token-based pattern matching to match using token attributes. If a string, will use medspaCy's
                RegexMatcher. If None, will use `literal` as the pattern for phrase matching. For more information, see
                https://spacy.io/usage/rule-based-matching.
            direction: The directionality or action of a modifier. This defines which part of a sentence a modifier will
                include as its scope. Entities within the scope will be considered to be modified.
                Valid values are:
                - "FORWARD": Scope will begin after the end of a modifier and move to the right
                - "BACKWARD": Scope will begin before the beginning of a modifier and move to the left
                - "BIDIRECTIONAL": Scope will expand on either side of a modifier
                - "TERMINATE": A special direction to limit any other modifiers if this phrase is in its scope. Example:
                    "no evidence of chf but there is pneumonia": "but" will prevent "no evidence of" from modifying
                    "pneumonia"
                - "PSEUDO": A special direction which will not modify any targets. This can be used for differentiating
                    superstrings of modifiers. Example: A modifier with literal="negative attitude" will prevent the
                    phrase "negative" in "She has a negative attitude about her treatment" from being extracted as a
                    modifier.
            on_match: An optional callback function or other callable which takes 4 arguments: `(matcher, doc, i,
                matches)`. For more information, see https://spacy.io/usage/rule-based-matching#on_match
            on_modifies: Callback function to run when building an edge between a target and a modifier. This allows
                specifying custom logic for allowing or preventing certain modifiers from modifying certain targets. The
                callable should take 3 arguments:
                    target: The spaCy Span from doc.ents (ie., 'Evidence of pneumonia')
                    modifier: The spaCy Span covered in a resulting modifier (ie., 'no evidence of')
                    span_between: The Span between the target and modifier in question.
                Should return either True or False. If returns False, then the modifier will not modify the target.
            allowed_types: A collection of target labels to allow a modifier to modify. If None, will apply to any type
                not specifically excluded in excluded_types. Only one of allowed_types and excluded_types can be used.
                An error will be thrown if both are not None.
            excluded_types: A collection of target labels which this modifier cannot modify. If None, will apply to all
                target types unless allowed_types is not None.
            max_scope: A number of tokens to explicitly limit the size of the modifier's scope. If None, the scope will
                include the entire sentence in the direction of `direction` and the entire sentence for "BIDIRECTIONAL".
                This is useful for requiring modifiers be very close to a concept in the text or for preventing long
                modifier ranges caused by sentence splitting problems.
            max_targets: The maximum number of targets which a modifier can modify. If None, will modify all targets in
                its scope.
            terminated_by: An optional collection of other modifier categories which will terminate the scope of this
                modifier. If None, only "TERMINATE" will do this. Example: if a ConTextRule defining "positive for" has
                terminated_by={"NEGATED_EXISTENCE"}, then in the sentence "positive for flu, negative for RSV", the
                positive modifier will modify "flu" but will be terminated by "negative for" and will not modify "RSV".
                This helps prevent multiple conflicting modifiers from distributing too far across a sentence.
            metadata: Optional dictionary of any extra metadata.
        """
        super().__init__(literal, category.upper(), pattern, on_match, metadata)
        self.on_modifies = on_modifies

        if allowed_types is not None and excluded_types is not None:
            raise ValueError(
                "A ConTextRule was instantiated with non-null values for both allowed_types and excluded_types. "
                "Only one of these can be non-null."
            )
        if allowed_types is not None:
            self.allowed_types = {label.upper() for label in allowed_types}
        else:
            self.allowed_types = None
        if excluded_types is not None:
            self.excluded_types = {label.upper() for label in excluded_types}
        else:
            self.excluded_types = None

        if max_targets is not None and max_targets <= 0:
            raise ValueError("max_targets must be >= 0 or None.")
        self.max_targets = max_targets
        if max_scope is not None and max_scope <= 0:
            raise ValueError("max_scope must be >= 0 or None.")
        self.max_scope = max_scope
        if terminated_by is None:
            terminated_by = set()
        else:
            if isinstance(terminated_by, str):
                raise ValueError(
                    f"terminated_by must be an iterable, such as a list or set, not {terminated_by}."
                )
            terminated_by = {string.upper() for string in terminated_by}

        self.terminated_by = terminated_by

        self.metadata = metadata

        if direction.upper() not in self._ALLOWED_DIRECTIONS:
            raise ValueError(
                "Direction {0} not recognized. Must be one of: {1}".format(
                    direction, self._ALLOWED_DIRECTIONS
                )
            )
        self.direction = direction.upper()

    @classmethod
    def from_json(cls, filepath) -> List[ConTextRule]:
        """
        Reads in a lexicon of modifiers from a JSON file under the key `context_rules`.

        Args:
            filepath: The .json file containing modifier rules. Must contain `context_rules` key containing the rule
                JSONs.

        Returns:
            A list of ConTextRules objects read from the JSON.
        """

        with open(filepath) as file:
            modifier_data = json.load(file)
        context_rules = []
        for data in modifier_data["context_rules"]:
            context_rules.append(ConTextRule.from_dict(data))
        return context_rules

    @classmethod
    def from_dict(cls, rule_dict) -> ConTextRule:
        """
        Reads a dictionary into a ConTextRule.

        Args:
            rule_dict: The dictionary to convert.

        Returns:
            The ConTextRule created from the dictionary.
        """
        keys = set(rule_dict.keys())
        invalid_keys = keys.difference(cls._ALLOWED_KEYS)
        if invalid_keys:
            msg = (
                "JSON object contains invalid keys: {0}.\n"
                "Must be one of: {1}".format(invalid_keys, cls._ALLOWED_KEYS)
            )
            raise ValueError(msg)
        rule = ConTextRule(**rule_dict)
        return rule

    def to_dict(self):
        """
        Converts ConTextItems to a python dictionary. Used when writing context rules to a json file.

        Returns:
            The dictionary containing the ConTextRule info.
        """
        rule_dict = {}
        for key in self._ALLOWED_KEYS:
            value = self.__dict__.get(key)
            if isinstance(value, set):
                value = list(value)
            if value is not None:
                rule_dict[key] = value
        return rule_dict

    @classmethod
    def to_json(cls, context_rules: List[ConTextRule], filepath: str):
        """Writes ConTextItems to a json file.

            Args:
            context_rules: a list of ContextRules that will be written to a file.
            filepath: the .json file to contain modifier rules
        """
        import json

        data = {"context_rules": [rule.to_dict() for rule in context_rules]}
        with open(filepath, "w") as file:
            json.dump(data, file, indent=4)

    def __repr__(self):
        return (
            f"ConTextRule(literal='{self.literal}', category='{self.category}', pattern={self.pattern}, "
            f"direction='{self.direction}')"
        )



"""
This module contains some simple functions that can be used as action or condition functions for postprocessing rules.
"""
from typing import Iterable, Union, Literal

from spacy.tokens import Span

from ..common.util import span_contains


def is_negated(span: Span) -> bool:
    """
    Returns whether a span is marked as negated.

    Args:
        span: The span to check.

    Returns:
        Whether the specified span has span._.is_negated set to True.
    """
    return span._.is_negated


def is_uncertain(span: Span) -> bool:
    """
    Returns whether a span is marked as uncertain.

    Args:
        span: The span to check.

    Returns:
        Whether the specified span has span._.is_uncertain set to True.
    """
    return span._.is_uncertain


def is_historical(span: Span) -> bool:
    """
    Returns whether a span is marked as historical.

    Args:
        span: The span to check.

    Returns:
        Whether the specified span has span._.is_historical set to True.
    """
    return span._.is_historical


def is_hypothetical(span: Span) -> bool:
    """
    Returns whether a span is marked as hypothetical.

    Args:
        span: The span to check.

    Returns:
        Whether the specified span has span._.is_hypothetical set to True.
    """
    return span._.is_hypothetical


def is_family(span: Span) -> bool:
    """
    Returns whether a span is marked as family.

    Args:
        span: The span to check.

    Returns:
        Whether the specified span has span._.is_family set to True.
    """
    return span._.is_family


def is_modified_by_category(span: Span, category: str) -> bool:
    """
    Returns whether a span is modified by a ConTextModifier of that type.

    Args:
        span: The span to check.
        category: The category to check whether a ConTextModifier of that type modifies the span.

    Returns:
        Whether the specified span has the specified modifier type.
    """
    for modifier in span._.modifiers:
        if modifier.category.upper() == category.upper():
            return True
    return False


def is_modified_by_text(
    span: Span, target: Union[str, Iterable[str]], regex: bool = True
) -> bool:
    """
    Returns whether a span is modified by a ConTextModifier with the specified text.

    Args:
        span: The span to check.
        target: The category to check whether a ConTextModifier with this text modifies the span.
        regex: If the `target` specified is a regex pattern. Default is True.

    Returns:
        Whether the specified span has the specified modifier type.
    """
    for modifier in span._.modifiers:
        if span_contains(modifier.span, target, regex):
            return True
    return False


def is_preceded_by(
    ent: Span, target: Union[str, Iterable[str]], window: int = 1
) -> bool:
    """
    Checks if an entity is preceded by a target word within a certain window. If any phrases in target are more than one
    token long, this may not capture it if window is smaller than the number of tokens. Case-insensitive.

    Args:
        ent: The span to check.
        target: A string or a collection of strings that will be searched for in the text preceding `ent`.
        window: The number of tokens to search for `target` preceding `ent`. Default is 1.

    Returns:
        Whether the entity specified is preceded by a target.
    """
    preceding_span = ent.doc[ent.start - window : ent.start]
    preceding_string = " ".join([token.text.lower() for token in preceding_span])
    if isinstance(target, str):
        return target.lower() in preceding_string
    for string in target:
        if string.lower() in preceding_string:
            return True
    return False


def is_followed_by(
    ent: Span, target: Union[str, Iterable[str]], window: int = 1
) -> bool:
    """
    Checks if an entity is followed by a target word within a certain window. If any phrases in target are more than one
    token long, this may not capture it if window is smaller than the number of tokens. Case-insensitive.

    Args:
        ent: The span to check.
        target: A string or a collection of strings that will be searched for in the text following `ent`.
        window: The number of tokens to search for `target` following `ent`. Default is 1.

    Returns:
        Whether the entity specified is followed by a target.
    """
    following_span = ent.doc[ent.end : ent.end + window]
    following_string = " ".join([token.text.lower() for token in following_span])
    if isinstance(target, str):
        return target.lower() in following_string
    for string in target:
        if string.lower() in following_string:
            return True
    return False


def ent_contains(
    ent: Span, target: Union[str, Iterable[str]], regex: bool = True
) -> bool:
    """
    Check if an entity occurs in the same sentence as another span of text. Case-insensitive.

    Args:
        ent: The span to check.
        target: A string or a collection of strings that will be searched inside `ent`.
        regex: If the `target` specified is a regex pattern. Default is True.

    Returns:
        Whether the target is contained in the ent.
    """
    return span_contains(ent, target, regex)


def sentence_contains(ent: Span, target: Union[str, Iterable[str]], regex=True) -> bool:
    """
    Check if an entity occurs in the same sentence as another span of text.

    Args:
        ent: The span to check.
        target: A string or a collection of strings that will be searched for in the text of the sentence containing
            `ent`.
        regex: If the `target` specified is a regex pattern. Default is True.
    """
    return span_contains(ent.sent, target, regex)


def remove_ent(
    ent: Span,
    i: int,
    input_type: Literal["ents", "group"] = "ents",
    span_group_name: str = "medspacy_spans",
):
    """
    Remove an entity at position [i] from doc.ents.

    Args:
        ent: The entity to remove.
        i: The index of `ent` in its source list.
        input_type: The source of the entity, either "ents" or "group".
        span_group_name: If `input_type` is "group", the name of the span group.
    """
    doc = ent.doc
    if input_type == "ents":
        doc.ents = doc.ents[:i] + doc.ents[i + 1 :]
    elif input_type == "group":
        t = list(doc.spans[span_group_name])
        doc.spans[span_group_name] = t[:i] + t[i + 1 :]
        # del doc.spans[span_group_name][i] # why does this not work?


def set_label(
    ent,
    i,
    input_type: Literal["ents", "group"] = "ents",
    span_group_name: str = "medspacy_spans",
    **kwargs
):
    """
    Creates a copy of the entity with a new label.

    WARNING: This is not fully safe, as spaCy does not allow modifying the label of a span. Instead, this creates a new
    copy and attempts to copy existing attributes, but this is not totally reliable.

    Args:
        ent: The entity to MODIFY.
        i: The index of `ent` in its source list.
        input_type: The source of the entity, either "ents" or "group".
        span_group_name: If `input_type` is "group", the name of the span group.
    """
    from spacy.tokens import Span

    new_ent = Span(ent.doc, ent.start, ent.end, label=kwargs["label"])
    # Copy any additional attributes
    # NOTE: This may not be complete and should be used with caution
    for (attr, values) in ent._.__dict__["_extensions"].items():
        setattr(new_ent._, attr, values[0])
    if input_type == "ents":
        if len(ent.doc.ents) == 1:
            ent.doc.ents = (new_ent,)
        else:
            ent.doc.ents = ent.doc.ents[:i] + (new_ent,) + ent.doc.ents[i + 1 :]
    else:
        if len(ent.doc.spans[span_group_name] == 1):
            ent.doc.spans[span_group_name] = (new_ent,)
        else:
            ent.doc.spans[span_group_name] = (
                ent.doc.spans[span_group_name][:i]
                + (new_ent,)
                + ent.doc.spans[span_group_name][i + 1 :]
            )


def set_negated(ent, i, value=True):
    """Set the value of ent._.is_negated to value."""
    ent._.is_negated = value


def set_uncertain(ent, i, value=True):
    """Set the value of ent._.is_uncertain to value."""
    ent._.is_uncertain = value


def set_historical(ent, i, value=True):
    """Set the value of ent._.is_historical to value."""
    ent._.is_historical = value


def set_hypothetical(ent, i, value=True):
    """Set the value of ent._.is_hypothetical to value."""
    ent._.is_hypothetical = value


def set_family(ent, i, value=True):
    "Set the value of ent._.is_family to value."
    ent._.is_hypothetical = value



import pytest

import sys
import medspacy
import spacy

from medspacy.target_matcher import TargetRule


class TestMedSpaCy:
    def test_default_build_pipe_names(self):
        enable, disable = medspacy.util._build_pipe_names(
            enable="default", disable=None
        )
        assert enable == {
            "medspacy_tokenizer",
            "medspacy_pyrush",
            "medspacy_target_matcher",
            "medspacy_context",
        }
        assert disable == set()

    def test_default_load(self):
        nlp = medspacy.load()
        expected_pipe_names = {
            "medspacy_pyrush",
            "medspacy_context",
            "medspacy_target_matcher",
        }
        assert set(nlp.pipe_names) == expected_pipe_names

    def test_load_enable(self):
        nlp = medspacy.load(
            medspacy_enable={"medspacy_target_matcher", "medspacy_sectionizer"}
        )
        assert len(nlp.pipeline) == 2
        assert set(nlp.pipe_names) == {
            "medspacy_target_matcher",
            "medspacy_sectionizer",
        }

    def test_nlp(self):
        nlp = medspacy.load()
        assert nlp("This is a sentence. So is this.")

    def test_load_disable(self):
        nlp = medspacy.load(medspacy_disable=["medspacy_context"])
        expected_pipe_names = {
            "medspacy_pyrush",
            "medspacy_target_matcher",
        }
        assert set(nlp.pipe_names) == expected_pipe_names

    def test_load_all_component_names(self):
        expected_pipe_names = {
            "medspacy_tokenizer",
            "medspacy_preprocessor",
            "medspacy_pyrush",
            "medspacy_target_matcher",
            # "medspacy_quickumls", # quickumls still not included by default due to install issues
            "medspacy_context",
            "medspacy_sectionizer",
            "medspacy_postprocessor",
            "medspacy_doc_consumer",
        }
        actual_pipe_names, _ = medspacy.util._build_pipe_names("all")
        assert expected_pipe_names == actual_pipe_names

    def test_load_all_components(self):
        full_pipe_names = [
            "medspacy_pyrush",
            "medspacy_target_matcher",
            "medspacy_context",
            "medspacy_sectionizer",
            "medspacy_postprocessor",
            "medspacy_doc_consumer",
        ]

        nlp = medspacy.load(medspacy_enable="all")
        assert nlp.pipe_names == full_pipe_names
        assert isinstance(nlp.tokenizer, medspacy.preprocess.Preprocessor)

    # def test_load_sci(self):
    #     # pip install https://s3-us-west-2.amazonaws.com/ai2-s2-scispacy/releases/v0.3.0/en_core_sci_sm-0.3.0.tar.gz
    #     assert medspacy.load("en_core_sci_sm")

    def test_load_rules(self):
        nlp = medspacy.load(load_rules=True)
        context = nlp.get_pipe("medspacy_context")
        assert context.rules

    def test_not_load_rules(self):
        nlp = medspacy.load(load_rules=False)
        context = nlp.get_pipe("medspacy_context")
        assert not context.rules

    def test_load_lang_model(self):
        # check that spacy pipeline components are still in the nlp object
        nlp = spacy.load("en_core_web_sm", exclude={"ner", "parser", "senter"})
        nlp = medspacy.load(nlp)
        assert {"tagger", "lemmatizer"}.intersection(set(nlp.pipe_names))

    def test_medspacy_tokenizer(self):
        default_tokenizer = spacy.blank("en").tokenizer
        custom_tokenizer = medspacy.load(
            medspacy_enable=["medspacy_tokenizer"]
        ).tokenizer

        text = r"Pt c\o n;v;d h\o chf+cp n/v/d"

        default_doc = default_tokenizer(text)
        medspacy_doc = custom_tokenizer(text)

        assert [token.text for token in default_doc] != [
            token.text for token in medspacy_doc
        ]

        # Check that some expected token boundries are generated
        joined_tokens = " ".join([token.text for token in medspacy_doc])
        assert "c \\ o" in joined_tokens
        assert "n / v / d" in joined_tokens
        assert "chf + cp" in joined_tokens

    def test_disable_medspacy_tokenizer(self):
        default_tokenizer = spacy.blank("en").tokenizer
        custom_tokenizer = medspacy.load(
            medspacy_disable=["medspacy_tokenizer"]
        ).tokenizer

        text = r"Pt c\o n;v;d h\o chf+cp n/v/d"

        default_doc = default_tokenizer(text)
        medspacy_doc = custom_tokenizer(text)

        assert [token.text for token in default_doc] == [
            token.text for token in medspacy_doc
        ]

    def test_medspacy_tokenizer_uppercase(self):
        custom_tokenizer = medspacy.load(
            medspacy_enable=["medspacy_tokenizer"]
        ).tokenizer

        # Issue 13: Ensure that uppercase tokens are not tokenized as each character
        # https://github.com/medspacy/medspacy/issues/13
        text = r"DO NOT BREAK ME UP"

        medspacy_doc = custom_tokenizer(text)

        tokens = [token.text for token in medspacy_doc]

        assert len(tokens) == 5

        # Check that some expected token boundries are generated
        joined_tokens = " ".join(tokens)
        assert "DO NOT BREAK ME UP" in joined_tokens
        assert "B R E A K" not in joined_tokens

    def test_medspacy_tokenizer_numerics(self):
        custom_tokenizer = medspacy.load(
            medspacy_enable=["medspacy_tokenizer"]
        ).tokenizer

        text = r"1.5 mg"

        medspacy_doc = custom_tokenizer(text)

        tokens = [token.text for token in medspacy_doc]

        assert len(tokens) == 2

        # Check that some expected token boundries are generated
        joined_tokens = " ".join(tokens)
        assert "1.5" in joined_tokens
        assert "1 . 5" not in joined_tokens


class TestMedSpaCyForRelease:
    @pytest.mark.parametrize(
        "language_model",
        [
            ("es_core_news_sm"),
            ("pl_core_news_sm"),
            ("de_core_news_sm"),
            ("xx_ent_wiki_sm"),
        ],
    )
    def test_multilingual_load(self, language_model):
        """
        Checks that we can instantiate the pipeline with different language backends
        """
        # Try instantiating the model
        try:
            nlp = medspacy.load(language_model, **{"disable": ["parser"]})
        # Except if you don't have the model downloaded
        except OSError:
            assert True
            return
        doc = nlp(
            "This is a very short piece of text that we want to use for testing. No patients were given type 2 diabetes "
            "as part of this test case. Podczas tego testu nie dano żadnemu pacjentowi cukrzycy typu drugiego."
        )
        assert doc

    def test_pipeline_initiate_with_span_groups(self):
        nlp2 = spacy.blank("en")
        matcher = nlp2.add_pipe(
            "medspacy_target_matcher", config={"result_type": "group"}
        )
        matcher.add(TargetRule("pneumonia", "CONDITION"))
        sectionizer = nlp2.add_pipe(
            "medspacy_sectionizer", config={"input_span_type": "group"}
        )

        doc = nlp2("Past Medical History: Pneumonia, stroke, and cancer")
        assert sectionizer.input_span_type == "group"
        assert len(doc.spans["medspacy_spans"]) > 0
        assert doc.spans["medspacy_spans"][0]._.is_historical is True



from setuptools import setup, find_packages

# read the contents of the README file
import os
from os import path

this_directory = path.abspath(path.dirname(__file__))

long_description = 'Medspacy is a clinical natural language processing toolkit'

# TODO: Fix this later for a richer description
#with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
#    long_description = f.read()

# function to recursively get files for resourcee
def package_files(directory):
    paths = []
    for (p, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", p, filename))
    return paths


# get all files recursively from /resources
resource_files = package_files("./resources")

# get all files recursively from /requirements
resource_files += package_files("./requirements")

# read requirements configuration
with open('requirements/requirements.txt') as f:
    required = f.read().splitlines()


def get_version():
    """Load the version from version.py, without importing it.
    This function assumes that the last line in the file contains a variable defining the
    version string with single quotes.
    """
    try:
        with open("medspacy/_version.py", "r") as f:
            version_value = f.read().split("\n")[0].split("=")[-1].strip()
            
            # remove either double-quotes or single-quotes
            version_value = version_value.replace("'", "")
            version_value = version_value.replace('"', "")
            
            #print(f'version value verbatim: {version_value}')
            
            return version_value
    except IOError:
        raise IOError


setup(
    name="medspacy",
    version=get_version(),
    description="Library for clinical NLP with spaCy.",
    author="medSpaCy",
    packages=find_packages(),
    install_requires=required,
    # TODO: Fix this later for a richer description
    #long_description=long_description,
    #long_description_content_type="text/markdown",
    package_data={"medspacy": resource_files},
    python_requires=">=3.8.0",
)



class DbReader:
    def __init__(self, db_conn, read_query, start=-1, end=-1, read_batch_size=-1):
        self.db = db_conn
        self.read_query = read_query
        self.start = start
        self.end = end
        self.batch_size = read_batch_size
        self.read_complete = False

    def read(self):
        if not self.read_complete:
            if self.batch_size < 0:
                result = self.db.read(self.read_query)
                self.read_complete = True
            else:
                result = self.db.read(
                    self.read_query.format(self.start, self.start + self.batch_size)
                )
                self.start += self.batch_size
                if self.start >= self.end:
                    self.read_complete = True
            return result
        else:
            return None

    def close(self):
        self.db.close()



from .doc_consumer import DocConsumer
from .db_connect import DbConnect
from .db_reader import DbReader
from .db_writer import DbWriter
from .pipeline import Pipeline

__all__ = ["DocConsumer", "DbConnect", "DbReader", "DbWriter", "Pipeline"]



from .context_rule import ConTextRule
from .context_modifier import ConTextModifier
from .context_graph import ConTextGraph

from .context import (
    ConText,
    DEFAULT_RULES_FILEPATH,
    DEFAULT_ATTRIBUTES,
)

__all__ = [
    "ConText",
    "ConTextRule",
    "DEFAULT_RULES_FILEPATH",
    "ConTextGraph",
    "ConTextModifier",
]



import spacy

from medspacy.io.db_connect import DbConnect
from medspacy.io.db_reader import DbReader
from medspacy.io.db_writer import DbWriter
from medspacy.io.doc_consumer import DocConsumer
from medspacy.io.pipeline import Pipeline

from medspacy.context import ConTextComponent
from medspacy.section_detection import Sectionizer

#####################################################
# SQL CONNECTION CONFIG
# Specify your sql connection
driver = "SQL Server"
server = "your_server"
db = "your_db"
user = ""
pwd = ""
#####################################################


#####################################################
# DATA DESTINATION CONFIG
# Specify where your output table is, how to interact
# iwith tables that already exist, and how many rows
# to write at a time.
destination_table = "your_new_table"
create_table = True
drop_existing = True
write_batch_size = 100
#####################################################


#####################################################
# DATA FORMAT CONFIG
# Specify the columns names and types in the output
# table. ID column should be at index 0.
#
# Note: This is not yet fully configurable. It is
# dependent on the configuration of the DocConsumer
# shown later.
row_type = "ent"  # 'ent' or 'section'
if row_type == "ent":
    # ENT OPTIONS:
    #   DEFAULT: text (str), start_char (str), end_char (str), label_ (str)
    #   CONTEXT: is_family (bool), is_negated (bool), is_uncertain (bool),
    #            is_hypothetical (bool), is_historical (bool)
    #   SECTION: section_title (str), section_parent (str)
    cols = [
        "id",
        "text",
        "start_char",
        "end_char",
        "label_",
        "is_hypothetical",
        "section_title",
    ]
    col_types = [
        "bigint",
        "varchar(100)",
        "int",
        "int",
        "varchar(50)",
        "bit",
        "varchar(10)",
    ]

elif row_type == "section":
    # SECTION OPTIONS: section_title (str), section_title_text (str),
    #                  section_title_start_char (int), section_title_end_char (int),
    #                  section_text (str), section_text_start_char (int),
    #                  section_text_end_char (int), section_parent (str)
    cols = [
        "id",
        "section_title",
        "section_title_text",
        "section_title_start_char",
        "section_title_end_char",
        "section_text",
        "section_text_start_char",
        "section_text_end_char",
        "section_parent",
    ]
    col_types = [
        "bigint",
        "varchar(200)",
        "varchar(max)",
        "int",
        "int",
        "varchar(max)",
        "int",
        "int",
        "varchar(200)",
    ]
#####################################################


#####################################################
# DATA READING CONFIGS
# Specify your data input configs. Query should be of
# format SELECT id, text FROM table. Batching info is
# OPTIONAL. If variables are not specified or if
# read_batch_size < 0, no batching will occur.
read_query = """select id, text
  from your_table
  where rowno>{0} and rowno<={1}"""
start = 0
end = 10
read_batch_size = 1
#####################################################


#####################################################
# NLP FACTORY
# initialize or call a method to produce your custom
# NLP pipeline here.
#
# NOTE: DocConsumer MUST be present at the end.
nlp = spacy.load("en_core_web_sm")
context = ConTextComponent(nlp)
sectionizer = Sectionizer(nlp, patterns=None)
sectionizer.add([{"section_title": "equals", "pattern": [{"LOWER": "=", "OP": "+"}]}])
consumer = DocConsumer(
    nlp, context=True, sectionizer=True
)  # DocConsumer has optional bool context and sectionizer
nlp.add_pipe(sectionizer)
nlp.add_pipe(context)
nlp.add_pipe(consumer)
#####################################################


#####################################################
# CREATING DB CONNECTIONS
db_read_conn = DbConnect(driver, server, db, user, pwd)
db_write_conn = DbConnect(driver, server, db, user, pwd)
#####################################################


#####################################################
# CREATING DB READ/WRITERS
db_reader = DbReader(
    db_read_conn, read_query, start, end, read_batch_size
)  # start, end, read_batch_size optional
db_writer = DbWriter(
    db_write_conn,
    destination_table,
    create_table,
    drop_existing,
    write_batch_size,
    cols,
    col_types,
)
#####################################################


#####################################################
# CREATING PIPELINE
pipeline = Pipeline(db_reader, db_writer, nlp, row_type)
#####################################################


#####################################################
# RUNNING NLP PIPELINE
pipeline.process()
#####################################################



"""
This module will contain helper functions and classes for common clinical processing tasks
which will be used in medspaCy's sectionizer.
"""
import re

from spacy.tokens import Doc

NEWLINE_PATTERN = r"[\n\r]+[\s]*$"


def is_start_line(idx: int, doc: Doc, pattern: re.Pattern) -> bool:
    """
    Check whether the token at idx occurs at the start of the line.

    Args:
        idx: The token index to check.
        doc: The doc to check in.
        pattern: The newline pattern to check with.

    Returns:
        Whether the token occurs at the start of a line.
    """
    # If it's the start of the doc, return True
    if idx == 0:
        return True
    # Otherwise, check if the preceding token ends with newlines
    preceding_text = doc[idx - 1].text_with_ws
    return pattern.search(preceding_text) is not None


def is_end_line(idx: int, doc: Doc, pattern: re.Pattern) -> bool:
    """
    Check whether the token at idx occurs at the end of the line.

    Args:
        idx: The token index to check.
        doc: The doc to check in.
        pattern: The newline pattern to check with.

    Returns:
        Whether the token occurs at the end of a line.
    """
    # If it's the end of the doc, return True
    if idx == len(doc) - 1:
        return True

    # Check if either the token has trailing newlines,
    # or if the next token is a newline
    text = doc[idx].text_with_ws
    if pattern.search(text) is not None:
        return True
    following_text = doc[idx + 1].text_with_ws
    return pattern.search(following_text) is not None
